﻿#region Using directives

using System;
using System.Collections;
using System.Text;
using System.IO.Ports;


#endregion

namespace MonsieurWJ.LibXMS2.Communicators.Basic {

	/// <summary>
	/// Capabilities provided by the <see cref="ICommunicator"/>
	/// </summary>
	[Flags]
	public enum Capabilities {

		/// <summary>
		/// Supports Address Book Retrieval
		/// </summary>
		SupportsPhoneBook,

		/// <summary>
		/// Supports Battery Status Retrieval
		/// </summary>
		SupportsBatteryStatus,

		/// <summary>
		/// Supports Signal Strength Retrieval
		/// </summary>
		SupportsSignalStrength
	}

	/// <summary>
	/// Message Folder to be used
	/// </summary>
	public enum MessageFolder {
		Unread,
		Read,
		Sent,
		Unsent,
		All
	}

	/// <summary>
	/// Message Store
	/// </summary>
	public enum MessageStore {
		Phone,
		SIM
	}

	/// <summary>
	/// Tells about the limits of the phonebook
	/// </summary>
	/// <remarks>
	/// Limits like no. of entries, max. character lengths of the entries.
	/// </remarks>
	public class PhoneBookLimits {
		private string ab;
		private int from;
		private int to;
		private int namelen;
		private int fonelen;

		public string Type {
			get {
				return this.ab;
			}
		}

		public int From {
			get {
				return this.from;
			}
		}

		public int To {
			get {
				return this.to;
			}
		}

		public int NameLength {
			get {
				return this.namelen;
			}
		}

		public int PhoneLength {
			get {
				return this.fonelen;
			}
		}

		internal PhoneBookLimits(string nm, int f, int t, int nl, int pl) {
			this.ab = nm;
			this.from = f;
			this.to = t;
			this.namelen = nl;
			this.fonelen = pl;
		}

		public override string ToString() {
			return ab + " [" + from + " - " + to + "] Name Length: " + namelen + ", Phone Length: " + fonelen;
		}
	}

	/// <summary>
	/// Communicator Interface
	/// </summary>
	/// <remarks>
	/// Communicator communicates with the modem and retrieves address books,
	/// messages etc.
	/// </remarks>
	public interface ICommunicator:IDisposable {

		/// <summary>
		/// Vendor of the modem this communicator is made to communicate
		/// with.
		/// </summary>
		/// <value>Vendor </value>
		string Vendor {
			get;
		}

		/// <summary>
		/// Gets Modem ID
		/// </summary>
		/// <value>Modem ID</value>
		string ModemID {
			get;
		}

		/// <summary>
		/// Gets Serial Number Identification
		/// </summary>
		/// <value>Serial Number Identificatio</value>
		string SerialNo {
			get;
		}

		/// <summary>
		/// Gets manufacturer Identification
		/// </summary>
		/// <value>Manufacturer Identification</value>
		string Manufacturer {
			get;
		}

		/// <summary>
		/// Gets Model Identification
		/// </summary>
		/// <value>Model Identification</value>
		string Model {
			get;
		}

		/// <summary>
		/// Gets Revision Identification
		/// </summary>
		/// <value>Revision Identification</value>
		string Revision {
			get;
		}

		/// <summary>
		/// Gets capabilities of this provider
		/// </summary>
		/// <value>Capabilities</value>
		Capabilities Capabilities {
			get;
		}

		/// <summary>
		/// Initializes this communicator
		/// </summary>
		/// <param name="port">Communication Port</param>
		void Initialize(ref SerialPort port);

		/// <summary>
		/// Sends the SMS SUBMIT Packet
		/// </summary>
		/// <param name="packet">SMS SUBMIT Packet</param>
		void Send(SmsSubmitPacket packet);

		/// <summary>
		/// Saves the SMS SUBMIT Packet
		/// </summary>
		/// <param name="packet">SMS SUBMIT Packet</param>
		void Save(SmsSubmitPacket packet);

		/// <summary>
		/// Sets preferred message store
		/// </summary>
		/// <param name="store">Type of store</param>
		void SetMessageStore(MessageStore store);

		/// <summary>
		/// Sets message center address
		/// </summary>
		/// <param name="msgCenter">Address of the message center</param>
		/// <param name="toa">Type of Address of the message center</param>
		void SetMessageCenter(string msgCenter, byte toa);

		/// <summary>
		/// Gets current message center address
		/// </summary>
		/// <returns>Message Center Address</returns>
		string GetMessageCenter();

		/// <summary>
		/// Gets Battery Charge Status of the cell phone
		/// </summary>
		/// <returns>Battery Charge Status (from 0 - 100, 0 
		/// means battery exhausted or not connected, -1 in case of error )
		/// </returns>
		int GetBatteryCharge();

		/// <summary>
		/// Gets signal strength of the cell phone with the cellular service provider
		/// </summary>
		/// <returns>Signal Strength (from 0 to 31 and 99 if not detectable, -1 if error)</returns>
		int GetSignalStrength();

		/// <summary>
		/// Reads entire contacts list
		/// </summary>
		/// <returns>A <see cref="PhoneBook"/> instance</returns>
		PhoneBook GetContacts();

		/// <summary>
		/// Sets current phone book
		/// </summary>
		/// <param name="phoneBook">Type of phone book</param>
		/// <exception cref="NoSuchPhoneBookException">
		/// If no such phone book is found.
		/// </exception>
		void SetPhoneBook(string phoneBook);

		/// <summary>
		/// Writes contact to the phone in the current phone book
		/// </summary>
		/// <param name="entry">Phone Book Entry</param>
		/// <remarks>
		/// An exception will raise if trying to write to a readonly phone book
		/// like. Dialled Calls, Missed Calls, Received Calls etc.<br/>
		/// Phone Book must be set prior to calling this function.
		/// Phone Book in the <paramref name="entry"/> is not used but <see cref="PhoneBookEntry.Location"/>
		/// is used.
		/// </remarks>
		void WriteContact(PhoneBookEntry entry);

		/// <summary>
		/// Deletes contact at specified index in the current phonebook
		/// </summary>
		/// <param name="id">Index</param>
		/// <exception cref="IndexOutOfBoundsException">
		/// If index is out of the range of phone
		/// </exception>
		void DeleteContact(int id);

		/// <summary>
		/// Deletes the message with specified index from the current message store
		/// </summary>
		/// <param name="id">Index</param>
		void DeleteMessage(int id);

		/// <summary>
		/// Retrieves messages from the cellphone
		/// </summary>
		/// <returns>An ICollection of <see cref="SMS"/></returns>
		ICollection GetMessages(MessageFolder folder);
	}
}
