﻿#region Using directives

using System;
using System.IO;
using System.Text;
using System.IO.Ports;
using System.Threading;
using System.Collections;

using System.Text.RegularExpressions;

using MonsieurWJ.LibXMS2.Communicators.Basic;

#endregion

namespace MonsieurWJ.LibXMS2.Communicators.Nokia {

	public class Kommunikator:ICommunicator {

		private SerialPort port = null;

		private Hashtable hashRanges = null;

		#region Modem Identification Stuff
		private string gmi = null;
		private string gmr = null;
		private string gmm = null;
		private string gsn = null;
		#endregion

		// Currently selected phone book's PhoneBookLimits
		private PhoneBookLimits phoneBook = null;

		// Regex for parsing CPBR strings
		// I don't know why I've used this maybe for a change in the parsing style
		private static Regex rex_CPBR = null;

		private ModemInputQueue queue = null;

		/// <summary>
		/// Static Constructor
		/// </summary>
		static Kommunikator() {
			rex_CPBR = new Regex("^" + Regex.Escape("+CPBR: (") + @"(?<from>[0-9]+)" + Regex.Escape("-") +
				"(?<to>[0-9]+)" + Regex.Escape("),") + @"(?<phonelen>[0-9]+)" + Regex.Escape(",") +
				"(?<contactlen>[0-9]+)$", RegexOptions.Compiled);
		}

		/// <summary>
		/// Constructs a new instance of this communicator
		/// </summary>
		public Kommunikator() {
			queue = new ModemInputQueue();
			hashRanges = new Hashtable();
		}

		#region ICommunicator Members
		public string Vendor {
			get {
				return "Nokia";
			}
		}

		public string ModemID {
			get { 
				return "Nokia";
			}
		}

		public Capabilities Capabilities {
			get {
				return Capabilities.SupportsPhoneBook | Capabilities.SupportsBatteryStatus |
					   Capabilities.SupportsSignalStrength;
			}
		}

		public string SerialNo {
			get { return this.gsn; }
		}

		public string Manufacturer {
			get { return this.gmi; }
		}

		public string Model {
			get { return this.gmm; }
		}

		public string Revision {
			get { return this.gmr; }
		}

		public void Initialize(ref System.IO.Ports.SerialPort port) {
			this.port = port;
			string reply = null;
			this.port.ReceivedEvent += new SerialReceivedEventHandler(queue.BytesRecieved);
			
			// Disable Echo
			this.port.Write("ATE0\r\n");
			reply = this.GetReply();

			// Disable Extended Errors
			this.port.Write("AT+CMEE=0\r\n");
			reply = this.GetReply();
			if (reply.IndexOf("OK") <0)
				throw new CrcException(reply);

			#region Phone Identification
			this.port.Write("AT+CGMI\r\n");
			reply = this.GetReply();
			if(reply.IndexOf("OK") >= 0) {
				string[] msg = reply.Split(new char[] { '\r','\n' }, 
					StringSplitOptions.RemoveEmptyEntries);
				if(!msg[0].StartsWith("OK"))
					this.gmi = msg[0];
			}

			this.port.Write("AT+GSN\r\n");
			reply = this.GetReply();
			if (reply.IndexOf("OK") >= 0) {
				string[] msg = reply.Split(new char[] { '\r', '\n' },
					StringSplitOptions.RemoveEmptyEntries);
				if (!msg[0].StartsWith("OK"))
					this.gsn = msg[0];
			}

			this.port.Write("AT+GMM\r\n");
			reply = this.GetReply();
			if (reply.IndexOf("OK") >= 0) {
				string[] msg = reply.Split(new char[] { '\r', '\n' },
					StringSplitOptions.RemoveEmptyEntries);
				if (!msg[0].StartsWith("OK"))
					this.gmm = msg[0];
			}

			this.port.Write("AT+GMR\r\n");
			reply = this.GetReply();
			if (reply.IndexOf("OK") >= 0) {
				string[] msg = reply.Split(new char[] { '\r', '\n' },
					StringSplitOptions.RemoveEmptyEntries);
				foreach (string x in msg) {
					if (x.StartsWith("OK"))
						break;
					this.gmr += x + "\r\n";
				}
			}
			#endregion

			#region Address Book Parameters Read
			this.port.Write("AT+CPBS=?\r\n");
			reply = this.GetReply();
			if (reply.IndexOf("OK") < 0)
				throw new CrcException(reply);
			else {

				string[] msg = reply.Split(new char[] { '\r', '\n' }, StringSplitOptions.RemoveEmptyEntries);
				string[] abs = null;
				int pl, nl, to, from;

				msg[0] = msg[0].Substring(msg[0].Trim().IndexOf("(")).TrimStart('(').TrimEnd(')');

				abs = msg[0].Split(',');

				for (int i = 0; i < abs.Length; i++) {
					abs[i] = abs[i].Trim('"');
					port.Write("AT+CPBS=\"" + abs[i] + "\"\r\n");
					reply = this.GetReply();

					if (reply.IndexOf("OK") < 0)
						continue;

					port.Write("AT+CPBR=?\r\n");
					reply = this.GetReply();

					if (reply.IndexOf("OK") < 0)
						continue;

					msg = reply.Split(new char[] { '\r', '\n' }, StringSplitOptions.RemoveEmptyEntries);

					Match m = rex_CPBR.Match(msg[0]);
					GroupCollection gc = m.Groups;

					if (gc.Count >= 5) {
						nl = Int32.Parse(gc["contactlen"].Value);
						pl = Int32.Parse(gc["phonelen"].Value);
						to = Int32.Parse(gc["to"].Value);
						from = Int32.Parse(gc["from"].Value);
						hashRanges.Add(abs[i], new PhoneBookLimits(abs[i], from, to, nl, pl));
//						System.Windows.Forms.MessageBox.Show(abs[i] + ": " + hashRanges[abs[i]]);
					}
				}
			}
			port.Write("AT+CSCS=\"UCS2\"\r\n");
			reply = this.GetReply();

			if (reply.IndexOf("OK") < 0)
				throw new CrcException(reply);

			this.phoneBook = (PhoneBookLimits)hashRanges[PhoneBookType.SIM];
			#endregion
		}

		public PhoneBook GetContacts() {
			string reply = null;
			PhoneBook pb = null;
			string[] msgs = null;
			string bookType = phoneBook.Type;

			// Switch to UCS2, so that we can retrieve all characters
			// of the messages. In Nokia 6610, the UCS2 messages appears
			// in the hex encoded bytes
			port.Write("AT+CSCS=\"UCS2\"\r\n");
			reply = GetReply();

			if (reply.IndexOf("OK") < 0)
				throw new CrcException(reply);

			port.Write("AT+CPBR=" + this.phoneBook.From + "," + this.phoneBook.To + "\r\n");
			reply = GetReply();

			if (reply.IndexOf("OK") < 0)
				throw new CrcException(reply);

			msgs = reply.Split(new char[] { '\r', '\n' }, StringSplitOptions.RemoveEmptyEntries);
			pb = new PhoneBook();

			foreach (string msg in msgs) {
				if (msg.StartsWith("OK"))
					break;
				pb.Add(Kommunikator.ParseEntry(msg, bookType));
			}

			return pb;
		}

		public void SetPhoneBook(string bookType) {
			PhoneBookLimits range = hashRanges[bookType] as PhoneBookLimits;
			string reply = null;

			if (range == null)
				throw new NoSuchPhoneBookException(bookType);

			port.Write("AT+CPBS=\"" + bookType + "\"\r\n");
			reply = this.GetReply();

			if (reply.IndexOf("OK") < 0)
				throw new CrcException(reply);

			this.phoneBook = range;
		}

		public void WriteContact(PhoneBookEntry entry) {

/*
			SIM Mode
			--------
			Max. Chars For UCS2 - 9
			 "     "    "  PCCP437 - 12
			 "     "    "  HEX - 9
			ME Mode
			-------
			Max. Chars For UCS2 - 50
			 "    "     "  PCCP437 - 50

			This concludes that ME stores in UCS2 whereas SIM stores in ______________
			format.

			TODO: Fill in the above blank
*/
			string name = entry.Name;
			string phone = entry.Phone;
			byte toa = entry.ToA;
			int loc = entry.Location;
			string reply = null;
			byte[] bytes = null;


			if (phoneBook.Type.Equals(PhoneBookType.Phone)) {
				if (entry.Name.Length > phoneBook.NameLength)
					throw new OutOfLimitsException("Name", phoneBook.NameLength, entry.Name.Length);
			} else if (phoneBook.Type.Equals(PhoneBookType.SIM)) {
				if (Utility.IsStringUnicode(entry.Name)) {
					if (entry.Name.Length > (phoneBook.NameLength - 3))
						throw new OutOfLimitsException("Name", phoneBook.NameLength, entry.Name.Length);
				} else {
					if (entry.Name.Length > phoneBook.NameLength)
						throw new OutOfLimitsException("Name", phoneBook.NameLength, entry.Name.Length);
				}
			}
			if(entry.Phone.Length > phoneBook.PhoneLength)
				throw new OutOfLimitsException("Phone", phoneBook.NameLength, entry.Name.Length);

			bytes = Encoding.BigEndianUnicode.GetBytes(name);
			name = HexEncoding.HEX.GetString(bytes);

			port.Write("AT+CPBW=" + loc + ",\"" + phone + "\"," + toa + ",\"" + name + "\"\r\n");
			reply = this.GetReply();
			if (reply.IndexOf("OK") < 0)
				throw new CrcException(reply);
		}

		public void DeleteContact(int id) {
			string reply = null;

			if (!(this.phoneBook.From <= id && id <= this.phoneBook.To))
				throw new IndexOutOfRangeException();

			port.Write("AT+CPBW=" + id + "\r\n");
			reply = this.GetReply();
			if (reply.IndexOf("OK") < 0)
				throw new CrcException(reply);
		}

		public void DeleteMessage(int id) {
			string reply = null;

			port.Write("AT+CMGD=" + id + "\r\n");
			reply = this.GetReply();

			if (reply.IndexOf("OK") < 0)
				throw new CrcException(reply);
		}

		public ICollection GetMessages(MessageFolder folder) {
			ArrayList list = new ArrayList();
			string reply = null; // this.GetReply();
			port.Write("AT+CMGL=" + Kommunikator.GetMessageFolder(folder) + "\r\n");

			reply = this.GetReply();
			string[] msg = reply.Split(new char[] { '\r','\n' }, 
				StringSplitOptions.RemoveEmptyEntries);
			int i = 1;
			int xs = 1;
			int id = 0;
			Encoding enc = HexEncoding.HEX;
			foreach (string x in msg) {
				if (x.StartsWith("OK"))
					break;
				try {
					if (i % 2 == 0) {
						object pak = SmsParser.Parse(enc.GetBytes(x));
						if (pak == null) {
							continue;
						}
						list.Add(new SMS(id, pak));
					} else {
						string[] m = x.Substring(x.IndexOf(":")+1).Split(',');
						id = Int32.Parse(m[0]);
					}
				} catch (Exception) {
				}
				i++;
			}
			return list;
		}

		public void SetMessageStore(MessageStore stor) {
			string store = Kommunikator.GetMessageStore(stor);
			string reply = null;

			port.Write("AT+CPMS="+store+","+store+",\"MT\"\r\n");
			reply = this.GetReply();
			if (reply.IndexOf("OK") < 0)
				throw new CrcException(reply);
		}

		public int GetBatteryCharge() {
			port.Write("AT+CBC\r\n");
			string reply = this.GetReply();
			

			if (reply.IndexOf("OK") >0) {
				string[] msg = reply.Split(new char[] { '\r', '\n' },
					StringSplitOptions.RemoveEmptyEntries);
				string charge = msg[0];

				// The first string contains Battery Charge Specification
				// +CBC: (0,1),(0-100)
				// +CBC: 0, 95
				// 95 is of use
				// 0 tells of battery status

				string sz = charge.Substring(charge.IndexOf(":") + 1).Trim();
				
				int bcs = Int32.Parse(sz.Substring(0, sz.IndexOf(',')));
				int bcl = Int32.Parse(sz.Substring(sz.IndexOf(',')+1).Trim());

				return bcl;
			}
			else
				throw new CrcException(reply);
		}

		public int GetSignalStrength() {
			port.Write("AT+CSQ\r\n");
			string reply = this.GetReply();

			if (reply.IndexOf("OK") > 0) {
				string[] msg = reply.Split(new char[] { '\r', '\n' },
					StringSplitOptions.RemoveEmptyEntries);
				string sig = msg[0];

				// The first string contains Battery Charge Specification
				// +CSQ: (0-31,99),(99)
				// +CSQ: 14, 99
				// 14 is of use

				int rssi = Int32.Parse(sig.Substring(sig.IndexOf(':') + 1, sig.IndexOf(',') - sig.IndexOf(':') - 1).Trim());
				return rssi;
			} else
				throw new CrcException(reply);
		}

		public void SetMessageCenter(string msgCenter, byte toa) {
			string reply = null;

			port.Write("AT+CSCA=\"" + msgCenter + "\"," + toa+"\r\n");
			reply = this.GetReply();

			if (reply.IndexOf("OK") < 0)
				throw new CrcException(reply);
		}

		public string GetMessageCenter() {
			port.Write("AT+CSCA?\r\n");
			string reply = this.GetReply();

			if (reply.IndexOf("OK") < 0)
				throw new CrcException(reply);

			string[] msg = reply.Split(new char[] { '\r', '\n' }, StringSplitOptions.RemoveEmptyEntries);
			msg[0] = msg[0].Substring(msg[0].IndexOf(":") + 1);
			string[] x = msg[0].Trim().Split(',');

			return x[0].Trim('"');
		}

		public void Send(SmsSubmitPacket packet) {
			byte[] bytes = packet.GetBytes();

			// Append Ctrl-Z of EOF character
			string sz = HexEncoding.HEX.GetString(bytes) + "\x1A";
			string reply = null;

			port.Write("AT+CMGS=" + (bytes.Length - 1) + "\r\n");

			// Retrieve immediate reply i.e. '>' character
			reply = this.GetReply(true);

			port.Write(sz);

			reply = this.GetReply(true);
			if (reply.IndexOf("ERROR") > 0)
				throw new CrcException(reply);
		}

		public void Save(SmsSubmitPacket packet) {

			if (packet.MessageLength > 160)
				throw new MessageTooLongException("Message Length: " + packet.MessageLength + ", Recommended: 160");

			byte[] bytes = packet.GetBytes();

			// Append Ctrl-Z or EOF character to the message bytes
			string sz = HexEncoding.HEX.GetString(bytes) + "\x1A";
			string reply = null;

			port.Write("AT+CMGW=" + (bytes.Length - 1) + "\r\n");
			reply = this.GetReply(true);

			port.Write(sz);

			reply = this.GetReply();

			if (reply.IndexOf("ERROR") > 0)
				throw new CrcException(reply);
		}

		#endregion

		#region IDisposable Members

		public void Dispose() {
			if (this.port.IsOpen)
				this.port.Close();
			this.port.Dispose();
			this.port = null;
		}

		#endregion

		#region Private Utility Stuff
		private string GetReply(bool i) {
			return queue.GetReply(i);
		}

		private string GetReply() {
			return queue.GetReply();
		}
		#endregion

		#region Utility Stuff for Nokia Cellfones, might be for others too...

		/// <summary>
		/// Gets Message Store in string format
		/// </summary>
		/// <param name="store">Message Store</param>
		/// <returns></returns>
		private static string GetMessageStore(MessageStore store) {
			switch (store) {
				case MessageStore.Phone:
					return "\"ME\"";
				case MessageStore.SIM:
					return "\"SM\"";
			}
			return null;
		}

		/// <summary>
		/// Gets numeric equivalent for a message folder
		/// </summary>
		/// <param name="folder">Message Folder</param>
		/// <returns>Numeric Equivalent of mesage folder</returns>
		private static int GetMessageFolder(MessageFolder folder) {
			switch (folder) {
				case MessageFolder.Unread:
					return 0;
				case MessageFolder.Read:
					return 1;
				case MessageFolder.Unsent:
					return 2;
				case MessageFolder.Sent:
					return 3;
				case MessageFolder.All:
					return 4;
			}
			return -1;
		}

		/// <summary>
		/// Parses the reply into phone book entry
		/// </summary>
		/// <param name="reply">Reply</param>
		/// <param name="pb">Type of Phonebook</param>
		/// <returns>An instance of <see cref="PhoneBookEntry"/></returns>
		private static PhoneBookEntry ParseEntry(string reply, string pb) {
			string x = reply.Substring(reply.IndexOf(':') + 1);
			string[] ab = x.Split(',');
			int ord;
			string name;
			string phone;
			byte toa;
			byte[] bytes = null;
			Encoding enc = Encoding.BigEndianUnicode;

			try {
				ord = Int32.Parse(ab[0].Trim());
				phone = ab[1].Trim().Trim('"');
				if (ab[2].Trim().Length > 0)
					toa = Byte.Parse(ab[2].Trim());
				else
					toa = 0;
				if (ab[3].Trim().Length > 0) {
					name = ab[3].Trim().Trim('"');
					bytes = HexEncoding.HEX.GetBytes(name);
					name = enc.GetString(bytes);
				} else
					name = "";

				return new PhoneBookEntry(ord, name, phone, toa, pb);
			} catch (Exception e) {
				System.Windows.Forms.MessageBox.Show("Exception: " + e.Message);
				return null;
			}
		}
		#endregion

	}
}
