using System;

namespace MonsieurWJ.LibXMS2 {

	/// <summary>
	/// Modem Error Type
	/// </summary>
	[Serializable]
	public enum ModemErrorType {

		/// <summary>
		/// No Error
		/// </summary>
		NoError,

		/// <summary>
		/// No Modem
		/// </summary>
		NoModem,

		/// <summary>
		/// Modem already in use
		/// </summary>
		AlreadyInUse,
		
		/// <summary>
		/// Timeout occured while I/O operation
		/// </summary>
		TimeoutOccured,

		/// <summary>
		/// Invalid state of modem
		/// </summary>
		InvalidState,

		/// <summary>
		/// Access is denied to the modem
		/// </summary>
		AccessDenied,
		
		/// <summary>
		/// Error response to a valid AT Command
		/// </summary>
		UnknownCommandSet,

		/// <summary>
		/// Error No communicator found for the specific modem
		/// </summary>
		NoCommunicatorFound,
	}

	/// <summary>
	/// Modem Exception
	/// </summary>
	/// <remarks>
	/// Represents errors occured while communicating with modem
	/// </remarks>
	public class ModemException : Exception {

		private ModemErrorType error = ModemErrorType.NoError;

		/// <summary>
		/// Gets or sets type of error occured
		/// </summary>
		/// <value>Type of error</value>
		public ModemErrorType ModemError {
			get {
				return this.error;
			}
			set {
				this.error = value;
			}
		}

		/// <summary>
		/// Constructs a new instance of ModemException
		/// </summary>
		/// <param name="error">Type of error</param>
		/// <param name="e">Inner Exception</param>
		public ModemException(ModemErrorType error, Exception e):base("Modem Error", e) {
			this.error = error;
		}

		/// <summary>
		/// Constructs a new instance of ModemException
		/// </summary>
		/// <param name="info">Serialization Information</param>
		/// <param name="context">Streaming Context</param>
		public ModemException(
		  System.Runtime.Serialization.SerializationInfo info,
		  System.Runtime.Serialization.StreamingContext context) : base( info, context ) { }
	}

	/// <summary>
	/// CRC Exception
	/// </summary>
	public class CrcException : Exception {
		/// <summary>
		/// Cellular Result Code Exception
		/// </summary>
		/// <param name="ex">Error Code</param>
		public CrcException(string ex):base(ex) {
		}
	}

	/// <summary>
	/// No such Address Book Exception
	/// </summary>
	public class NoSuchPhoneBookException : Exception {
		public NoSuchPhoneBookException(string ab):base("No such phone book: "+ab) {
		}
	}

	/// <summary>
	/// Short Message Too Long
	/// </summary>
	public class MessageTooLongException : Exception {

		/// <summary>
		/// Constructs a new instance of <see cref="MessageTooLongException"/>
		/// </summary>
		/// <param name="message">Exception Message</param>
		public MessageTooLongException(string message) : base( message ) { }
	}

	/// <summary>
	/// Out of Limits Exception
	/// </summary>
	/// <remarks>
	/// Thrown when any parameter in the phone book goes out of limits
	/// </remarks>
	public class OutOfLimitsException : Exception {

		/// <summary>
		/// Constructs a new instance of <see cref="OutOfLimitsException"/>
		/// </summary>
		/// <param name="expected">Expected Limit</param>
		/// <param name="current">Exceptional Value</param>
		/// <param name="p">Name of the parameter</param>
		public OutOfLimitsException(string p, int expected, int current) : base( p+" -- Expected Limit: "+expected+", Current: "+current) { }
	}
}