﻿#region Using directives

using System;
using System.Collections;
using System.Text;

#endregion

namespace MonsieurWJ.LibXMS2 {

	/// <summary>
	/// GSM Encoding
	/// </summary>
	/// <remarks>
	/// TODO: Add support for &lt;ESC&gt; &lt;x&gt; strings i.e. escape codes
	/// </remarks>
	public class GsmEncoding : Encoding {

		static GsmEncoder encoder = null;
		static GsmDecoder decoder = null;
		static GsmEncoding encoding = null;

		static GsmEncoding() {
			GsmEncoding.encoder = new GsmEncoder();
			GsmEncoding.decoder = new GsmDecoder();
			GsmEncoding.encoding = new GsmEncoding();
		}

		public static GsmEncoding GSM {
			get {
				return encoding;
			}
		}

		public override string EncodingName {
			get { 
				return "GSM"; 
			}
		}

		public override int GetByteCount(char[] chars, int index, int count) {
			return GsmEncoding.encoder.GetByteCount(chars, index, count, true);
		}

		public override int GetBytes(char[] chars, int charIndex, int charCount, byte[] bytes, int byteIndex) {
			return GsmEncoding.encoder.GetBytes(chars, charIndex, charCount, bytes, byteIndex, true);
		}

		public override int GetCharCount(byte[] bytes, int index, int count) {
			return GsmEncoding.decoder.GetCharCount(bytes, index, count);
		}

		public override int GetChars(byte[] bytes, int byteIndex, int byteCount, char[] chars, int charIndex) {
			return GsmEncoding.decoder.GetChars(bytes, byteIndex, byteCount, chars, charIndex);
		}

		/// <summary>
		/// Gets actual (not really actual) byte count.
		/// </summary>
		/// <param name="charCount">Character Count</param>
		/// <returns>Byte Count</returns>
		public int GetByteCount(int charCount) {
			int i = charCount * 7;
			return ((i % 8) == 0 ? (i >> 3) : (i >> 3) + 1);
		}

		/// <summary>
		/// Gets maximum byte count
		/// </summary>
		/// <param name="charCount">Character Count</param>
		/// <returns>Byte Count</returns>
		public override int GetMaxByteCount(int charCount) {
			return ((charCount * 7) >> 3) + 1;
		}

		/// <summary>
		/// Gets Maximum Character Count
		/// </summary>
		/// <param name="byteCount">Byte Count</param>
		/// <returns>Character Count</returns>
		public override int GetMaxCharCount(int byteCount) {
			return (byteCount << 3)/7;
		}
	}

	/// <summary>
	/// GSM Decoder
	/// </summary>
	/// <remarks>
	/// <para>
	/// After GSM Decoder decodes the bytes into the string, the string has to be unescaped. i.e. all
	/// escape codes present in the string should be removed to get the special characters.
	/// To unescape the string, use <see cref="GsmHelper.Unescape"/> method.</para>
	/// <para>
	/// <strong>Note:</strong> After decoding substring the resultant string output into your expected one.
	/// Why this has to be done because, there are some cases where addition of a byte could result into
	/// two GSM characters. like e.g. take 55 bytes, in which only 62 characters can be encoded. But in 56
	/// bytes, 63 or 64 characters can be encoded. So it has to be decided, what user expects, thats why
	/// <see cref="String.Substring"/> has to be done after decoding.
	/// </para>
	/// </remarks>
	public class GsmDecoder : Decoder {

		private static int[] arr = new int[] { 0x7F, 0x3F, 0x1F, 0x0F, 0x07, 0x03, 0x01, 0x00 };

		private static int[] larr = new int[] { 0, 0x80, 0xC0, 0xE0, 0xF0, 0xF8, 0xFC, 0xFE, 0xFF };

		/// <summary>
		/// GSM Character Set
		/// </summary>
		public static char[] charset = new char[] {
			'@',
			'\xA3', /* Pound Sign */
			'$',
			'\xA5', /* Yen Sign */
			'\xE8', /* ISO 232 */
			'\xE9', /* ISO 233 */
			'\xF9', /* ISO 249 */
			'\xEC', /* ISO 236 */
			'\xF2', /* ISO 242 */
			'\xC7', /* cedilla 199 */
			'\n', /* LF */
			'\xD8', /* ISO 216 */
			'\xF2', /* ISO 242 */
			'\r', /* CR */
			'\xC5', /* Angstrom ISO 197 */
			'\xE5', /* Mini Angstrom ISO 229 */
			'\x394', /* Capital Delta */
			'_', /* Underscore Mini Underbar */
			'\x3A6', /* Capital PHI */
			'\x393', /* Capital GAMMA */
			'\x39B', /* Capital LAMBDA */
			'\x3A9', /* Capital OMEGA */
			'\x3A0', /* Capital PI */
			'\x3A8', /* Capital PSI */
			'\x3A3', /* Capital SIGMA */
			'\x398', /* Capital THETA */
			'\x39E', /* Capital XI */
			'\x1B', /* ESC */
			'\xC6', /* Capital AE 198 */
			'\xE6', /* Small AE 198 */
			'\xDF', /* Sharp S German 223 */
			'\xC9', /* Grave E 201 */
			' ',
			'!',
			'"',
			'#',
			'\xA4', /* Currency Symbol */
			'%',
			'&',
			'\'',
			'(',
			')',
			'*',
			'+',
			',',
			'-',
			'.',
			'/',
			'0',
			'1',
			'2',
			'3',
			'4',
			'5',
			'6',
			'7',
			'8',
			'9',
			':',
			';',
			'<',
			'=',
			'>',
		    '?',
			'\xA1', /* Opposite Exclamation Mark - 161 */
			'A',
			'B',
			'C',
			'D',
			'E',
			'F',
			'G',
			'H',
			'I',
			'J',
			'K',
			'L',
			'M',
			'N',
			'O',
			'P',
			'Q',
			'R',
			'S',
			'T',
			'U',
			'V',
			'W',
			'X',
			'Y',
			'Z',
			'\xC4', /* 196 */
			'\xD6', /* 214 */
			'\xD1', /* 209 */
			'\xDC', /* 220 */
		    '\xA7', /* 167 */
			'\xBF', /* Opposite Question Mark 191 */
			'a',
			'b',
			'c',
			'd',
			'e',
			'f',
			'g',
			'h',
			'i',
			'j',
			'k',
			'l',
			'm',
			'n',
			'o',
			'p',
			'q',
			'r',
			's',
			't',
			'u',
			'v',
			'w',
			'x',
			'y',
			'z',
			'\xE4', /* a 228 */
			'\xF6', /* o 246 */
			'\xF1', /* n 241 */
			'\xFC', /* u 252 */
			'\xE0'  /* u 224 */
		};

		public override int GetCharCount(byte[] bytes, int index, int count) {
			return (count << 3) / 7;

		}

		public override int GetChars(byte[] bytes, int byteIndex, int byteCount, char[] chars, int charIndex) {
			int len = (byteCount << 3) / 7;
			int j = 0;
			int o = 0;
			int x = 0;
			byte b = 0;

			for (int i = 0; i <= byteCount; i++) {
				o = i + byteIndex;

				if (o == byteCount)
					b = 0;
				else
					b = bytes[o];

				chars[x] = charset[((b & arr[j]) << j) | (i - 1 < 0 ? 0 : (bytes[o - 1] & larr[j]) >> (8 - j))]; //LBIT(array[o - 1], j))];
				j++;
				x++;
				if (x >= len)
					break;
				if (j > 7) {
					j = 1;
					chars[x] = charset[((b & arr[j - 1]) << (j - 1))];
					x++;
					if (x >= len)
						break;
				}
			}

			return x;
		}
	}

	/// <summary>
	/// GSM Encoder
	/// </summary>
	/// <remarks>
	/// String to be encoded should not be escaped prior to encoding. The escaping is builtin in to this
	/// encoder.
	/// </remarks>
	public class GsmEncoder : Encoder {

		private static Hashtable hMap = null;

		static GsmEncoder() {
			char l = (char)GsmDecoder.charset.Length;
			hMap = new Hashtable();

			for (char i = '\0'; i < l; i++)
				hMap[GsmDecoder.charset[i]] = i;
		}

		private int power(int x) {
			int i = 0;
			int v = 0;

			for (i = 1; i <= x; i++)
				v |= 1 << (i - 1);

			return v;
		}

		private static string MapString(char[] c, int index, int count) {
			char[] ch = new char[count];
			object o;
			Array.Copy(c, index, ch, 0, count);
			for (int i = 0; i < c.Length; i++) {
				o = hMap[c[i]];
				if (o != null)
					ch[i] = (char)o;
			}

			return new string(ch);
		}

		public override int GetByteCount(char[] chars, int index, int count, bool flush) {
			char[] c = GsmHelper.Escape(GsmEncoder.MapString(chars, index, count)).ToCharArray();
			int i = c.Length * 7;
			return ((i % 8) == 0 ? (i >> 3) : (i >> 3) + 1);
		}

		public override int GetBytes(char[] chars, int charIndex, int charCount, byte[] bytes, int byteIndex, bool flush) {
		    int i = 1;
			int k = 0;
			int j = 7;
			int l = 0;
			int x = 0;
			int m = 0;
			char[] c = GsmHelper.Escape(GsmEncoder.MapString(chars, charIndex, charCount)).ToCharArray();
			int len = GetByteCount(c, 0, c.Length, flush);
			byte[] b = new byte[len + 1];

			while (l < c.Length) {
				b[k] = (byte)((c[l]  & (power(j) << (7-j)))>> (i - 1));

				if (l+1 < c.Length)
					b[k] |= (byte)((c[l + 1] & power(i)) << j);

				j--;
				i++;

		        if(j == 0) {
					j = 7;
					l++;
				}
				if(i == 8)
					i = 1;
				l++;
				k++;
			}
			Array.Copy(b, 0, bytes, byteIndex, len);

			return l;
		}
	}

	/// <summary>
	/// GSM Encoding Helper
	/// </summary>
	public class GsmHelper {

		private static Hashtable hUnescape = null;
		private static Hashtable hEscape = null;

		/// <summary>
		/// Meta Class Constructor
		/// </summary>
		static GsmHelper() {
			hUnescape = new Hashtable();
			hEscape = new Hashtable();

			hUnescape[GsmDecoder.charset[0xA]] = '\xC';  /* Form Feed - 12  ---> 10 */
			hUnescape[GsmDecoder.charset[0x14]] = '^'; /* Caret - 94  --> 20 */
			hUnescape[GsmDecoder.charset[0x28]] = '{'; /* Left Curly Brace - 123 --> 40 */
			hUnescape[GsmDecoder.charset[0x29]] = '}';  /* Right Curly Brace - 125 --> 41 */
			hUnescape[GsmDecoder.charset[0x2F]] = '\\';  /* Forward Slash - 92 --> 47 */
			hUnescape[GsmDecoder.charset[0x3C]] = '[';  /* Left Square Brace - 91 --> 60 */
			hUnescape[GsmDecoder.charset[0x3D]] = '~';  /* Tilde - 126 --> 61 */
			hUnescape[GsmDecoder.charset[0x3E]] = ']';  /* Right Square Brace - 93 --> 62 */
			hUnescape[GsmDecoder.charset[0x40]] = '|';  /* Pipe - 124 --> 64 */
			hUnescape[GsmDecoder.charset[0x65]] = '\xA4';  /* Currency Symbol - 164 --> 101 */

			hEscape['\xC'] = '\xA';  /* Form Feed - 12  ---> 10 */
			hEscape['^'] = '\x14'; /* Caret - 94  --> 20 */
			hEscape['{'] = '\x28'; /* Left Curly Brace - 123 --> 40 */
			hEscape['}'] = '\x29';  /* Right Curly Brace - 125 --> 41 */
			hEscape['\\'] = '\x2F';  /* Forward Slash - 92 --> 47 */
			hEscape['['] = '\x3C';  /* Left Square Brace - 91 --> 60 */
			hEscape['~'] = '\x3D';  /* Tilde - 126 --> 61 */
			hEscape[']'] = '\x3E';  /* Right Square Brace - 93 --> 62 */
			hEscape['|'] = '\x40';  /* Pipe - 124 --> 64 */
		}

		/// <summary>
		/// No instantiation
		/// </summary>
		private GsmHelper() {
		}

		/// <summary>
		/// Escapes string for the GSM Character Set Special Characters
		/// </summary>
		/// <remarks>
		/// Escaping of text is built in to the encoder.
		/// </remarks>
		/// <param name="t">String to be escaped</param>
		/// <returns>Escaped String</returns>
		public static string Escape(string t) {
			foreach (char c in hEscape.Keys) {
				t = t.Replace(c.ToString(), "\x1b" + hEscape[c].ToString());
			}

			return t;
		}

		/// <summary>
		/// Unescapes string to get GSM Character Set Special Characters
		/// </summary>
		/// <remarks>
		/// Unescaping of text is not built in to the decoder. Unescaping is not using the GSM escape codes
		/// instead it is using the GSM character set equivalent codes. For more information, see source code.
		/// </remarks>
		/// <param name="t">String to be unescaped</param>
		/// <returns>Unescaped String</returns>
		public static string Unescape(string t) {
			foreach (char c in hUnescape.Keys) {
				t = t.Replace("\x1b" + c.ToString(), ""+hUnescape[c]);
			}

			return t;
		}
	}
}
