﻿#region Using directives

using System;
using System.Globalization;
using System.Collections.Generic;
using System.Text;

#endregion

namespace MonsieurWJ.LibXMS2 {

	#region Required Types
	/// <summary>
	/// Message Type
	/// </summary>
	/// <remarks>
	/// First Octet in the SMS &quot;PDU Packet&quot;
	/// </remarks>
	public enum MessageType:byte {
		/// <summary>
		/// SMS DELIVER message
		/// </summary>
		Deliver = 0,

		/// <summary>
		/// SMS SUBMIT message
		/// </summary>
		Submit = 1,

		/// <summary>
		/// SMS STATUS REPORT message
		/// </summary>
		StatusReport = 2
	}

	/// <summary>
	/// Class of the message
	/// </summary>
	public enum MessageClass : byte {

		/// <summary>
		/// Class 0 Message or Flash SMS, or Immediate Display
		/// </summary>
		Class0 = 0,

		/// <summary>
		/// Class 1 Message or ME Specific
		/// </summary>
		Class1,

		/// <summary>
		/// Class 2 Message or SM Specific
		/// </summary>
		Class2,

		/// <summary>
		/// Class 3 Message or TE Specific
		/// </summary>
		Class3
	}

	/// <summary>
	/// SMS Packet Flags
	/// </summary>
	[Flags]
	public enum PacketFlags {

		/// <summary>
		/// Use default SMSC
		/// </summary>
		UseSystemSMSC,

		/// <summary>
		/// Set Message Reference No.
		/// </summary>
		SetMsgRef
	}

	/// <summary>
	/// Type of Encoding
	/// </summary>
	public enum EncodingType {

		/// <summary>
		/// GSM Encoding
		/// </summary>
		GSM=0,

		/// <summary>
		/// 8 Bit Encoding
		/// </summary>
		OCTET=1,

		/// <summary>
		/// UCS 2 or 16 bit Unicode Encoding
		/// </summary>
		UCS2=2,

		/// <summary>
		/// Reserved
		/// </summary>
		Reserved=3
	}

	/// <summary>
	/// Type of Address
	/// </summary>
	public enum AddressType : byte {
		
		/// <summary>
		/// Address is Generic i.e. Unknown Phone Number Format.
		/// </summary>
		Generic = (byte)0x81,

		/// <summary>
		/// An International Number.
		/// </summary>
		International = (byte)0x91,

		/// <summary>
		/// A National Number
		/// </summary>
		National = (byte)0xA1,

		/// <summary>
		/// A number encoded in GSM alphabet.
		/// </summary>
		/// <remarks>
		/// Usually, this number is used by phone service providers
		/// </remarks>
		GSM = (byte)0xD0
	}

	/// <summary>
	/// Type of Validity Period Specified
	/// </summary>
	public enum ValidityType:byte {

		/// <summary>
		/// Not Present (0 octet)
		/// </summary>
		NotPresent = 0,

		/// <summary>
		/// Enhanced Format. (7 Octets)
		/// </summary>
		/// <remarks>
		/// See ETSI 03.40. <strong style="color: red">Not Yet Implemented</strong>
		/// </remarks>
		Enhanced,

		/// <summary>
		/// Relative Octet (1 octet)
		/// </summary>
		Relative,

		/// <summary>
		/// Absolute Format (given as <see cref="DateTime"/>)
		/// </summary>
		/// <remarks>
		/// SCTS style, <see cref="DateTime"/> when Short Message expires
		/// </remarks>
		Absolute
	}
	#endregion

	#region SMS Parser
	/// <summary>
	/// SMS Parser
	/// </summary>
	public class SmsParser {

		/// <summary>
		/// Make it uninstantiable
		/// </summary>
		private SmsParser() {
		}

		/// <summary>
		/// Parses SMS
		/// </summary>
		/// <param name="bytes">An array of bytes</param>
		/// <returns>An instance of one of the SMS packets if successful</returns>
		public static object Parse(byte[] bytes) {

			// Magic Field is TP-MTI ie. Message Type Indicator
			// First two bits (0 and 1) of the first octet
			switch (bytes[bytes[0] + 1] & 3) {
				case (byte)MessageType.Deliver:
					return SmsDeliverPacket.Parse(bytes);
				case (byte)MessageType.Submit:
					return SmsSubmitPacket.Parse(bytes);
				case (byte)MessageType.StatusReport:
					return SmsStatusReportPacket.Parse(bytes);
			}

			return null;
		}
	}
	#endregion

	#region Different SMS Packets

	/// <summary>
	/// A generic way to represent a Short Message
	/// </summary>
	public class SMS {
		object packet;
		int id;

		/// <summary>
		/// Gets SMS Packet
		/// </summary>
		/// <value>SMS Packet</value>
		public object Packet {
			get {
				return packet;
			}
		}

		/// <summary>
		/// Gets the index of the message in the message folder
		/// </summary>
		/// <value></value>
		public int ID {
			get {
				return this.id;
			}
		}

		/// <summary>
		/// Constructs a new instance of <see cref="SMS"/>.
		/// </summary>
		/// <param name="id">Index of the message in the message folder</param>
		/// <param name="packet">SMS Packet</param>
		public SMS(int id, object packet) {
			this.id = id;
			this.packet = packet;
		}
	}

	/// <summary>
	/// SMS DELIVER packet
	/// </summary>
	public class SmsDeliverPacket {

		#region Private members
		private byte len_smsc;
		private byte toa_smsc;
		private byte[] smsc_no = null;

		private byte first_octet;
		
		private byte len_sender;
		private byte toa_sender;
		private byte[] sender_no = null;
		
		private byte pid;
		private byte dcs;
		
		private byte[] scts = null;
		
		private byte udl;
		private byte[] msg = null;

		private string text = null;
		private string sender = null;
		private string smsc = null;

		private DateTime timestamp;

		private byte msg_len = 0;
		#endregion

		#region Constant Stuff...
		private const int TP_MMS = 2;
		private const int TP_SRI = 5;
		private const int TP_UDHI = 6;
		private const int TP_RP = 7;
		#endregion

		/// <summary>
		/// Gets whether reply path exists for this SM
		/// </summary>
		/// <value>Is Reply Path Exists</value>
		public bool ReplyPathExists {
			get {
				return ((this.first_octet & (1 << SmsDeliverPacket.TP_RP)) == (1 << SmsDeliverPacket.TP_RP)) ? true : false;
			}
		}

		/// <summary>
		/// Gets whether user data header exists
		/// </summary>
		/// <value>True if user data header exists else false</value>
		public bool UserDataHeaderExists {
			get {
				return ((this.first_octet & (1 << SmsDeliverPacket.TP_UDHI)) == (1 << SmsDeliverPacket.TP_UDHI)) ? true : false;
			}
		}

		/// <summary>
		/// Gets whether delivery report is requested for this SMS
		/// </summary>
		/// <value>Is delivery report requested</value>
		public bool WantsDeliveryReport {
			get {
				return ((this.first_octet & (1 << SmsDeliverPacket.TP_SRI)) == (1 << SmsDeliverPacket.TP_SRI)) ? true : false;
			}
		}

		/// <summary>
		/// Gets whether there are any more messages
		/// </summary>
		/// <value>Is there any more message</value>
		public bool MoreMessages {
			get {
				return ((this.first_octet & (1 << SmsDeliverPacket.TP_MMS)) == (1 << SmsDeliverPacket.TP_MMS)) ? false : true;
			}
		}

		/// <summary>
		/// Gets timestamp on this message
		/// </summary>
		/// <value>Timestamp on this message</value>
		public DateTime Timestamp {
			get {
				return timestamp;
			}
		}

		/// <summary>
		/// Gets UTC Offset for the timestamp's timezone
		/// </summary>
		/// <value>UTC Offset of the timestamp</value>
		public TimeSpan TimeZone {
			get {
				return Utility.QuartersBcdToUtc(this.scts[6]);
			}
		}

		/// <summary>
		/// Gets SMS Text
		/// </summary>
		/// <value>SMS Text</value>
		public string MessageText {
			get {
				return text;
			}
		}

		/// <summary>
		/// Gets Sender Number
		/// </summary>
		/// <value>Sender Number</value>
		public string OriginatingAddress {
			get {
				return sender;
			}
		}

		/// <summary>
		/// Gets Type of Address of SMSC
		/// </summary>
		/// <value>Type of Address</value>
		public byte SMSC_ToA {
			get {
				return toa_smsc;
			}
		}

		/// <summary>
		/// Gets Type of Address of Sender
		/// </summary>
		/// <value>Type of Address</value>
		public byte OriginatingAddress_ToA {
			get {
				return toa_sender;
			}
		}

		/// <summary>
		/// Gets SMSC (Short Message Service Center or SMS Center) No.
		/// </summary>
		/// <value>SMS Center No.</value>
		public string SMSC {
			get {
				return smsc;
			}
		}

		/// <summary>
		/// Gets Class of the Message
		/// </summary>
		/// <value>Message Class</value>
		public MessageClass MessageClass {
			get {
				return Utility.GetMessageClass(this.dcs);
			}
		}

		/// <summary>
		/// Gets Class of the Message
		/// </summary>
		/// <value>Message Class</value>
		public EncodingType MessageEncoding {
			get {
				return Utility.GetEncoding(this.dcs);
			}
		}

		/// <summary>
		/// Constructs a new instance of the SmsDeliverPacket
		/// </summary>
		protected SmsDeliverPacket() {
			this.len_smsc = 0;
			this.toa_smsc = 0;
			this.first_octet = 0;
			this.pid = 0;
			this.dcs = 0;
			this.udl = 0;
			this.toa_sender = 0;
			this.len_sender = 0;
		}

		/// <summary>
		/// Process this SMS packet
		/// </summary>
		private void Process() {

			// Find the Timestamp
			timestamp = Utility.BytesToTimestamp(this.scts);

			this.sender = Utility.BytesToPhone(this.toa_sender, this.len_sender, 0, this.sender_no);
			this.smsc = Utility.BytesToPhone(this.toa_smsc, this.len_smsc, 1, this.smsc_no);

			switch (Utility.GetEncoding(this.dcs)) {
				case EncodingType.GSM:
					this.text = GsmEncoding.GSM.GetString(msg);
					// Substring the message to the User Data Length
					this.text = text.Substring(0, this.udl);
					this.text = GsmHelper.Unescape(this.text);
					break;
				case EncodingType.OCTET:
					this.text = Encoding.ASCII.GetString(msg, 0, msg.Length);
					break;
				case EncodingType.UCS2:
					this.text = Encoding.BigEndianUnicode.GetString(msg, 0, msg.Length);
					break;
			}
		}

		/// <summary>
		/// Creates an instance of the <see cref="SmsDeliverPacket"/> by parsing the
		/// SMS DELIVER packet bytes.
		/// </summary>
		/// <param name="bytes">SMS DELIVER packet bytes</param>
		/// <returns>An instance of <see cref="SmsDeliverPacket"/> if successful,
		/// otherwise null</returns>
		public static SmsDeliverPacket Parse(byte[] bytes) {
			SmsDeliverPacket packet = new SmsDeliverPacket();
			int idx = 0;
			int ml = 0;

			packet.len_smsc = bytes[idx];

			if (packet.len_smsc != 0) {
				packet.toa_smsc = bytes[++idx];
				packet.smsc_no = new byte[packet.len_smsc - 1];
				Array.Copy(bytes, ++idx, packet.smsc_no, 0, packet.smsc_no.Length);
				idx += packet.smsc_no.Length - 1;
				Utility.Swap(packet.smsc_no, 0, packet.smsc_no.Length);
			}

			packet.first_octet = bytes[++idx];
			packet.len_sender = bytes[++idx];
			if (packet.len_sender != 0) {
				packet.toa_sender = bytes[++idx];
				packet.sender_no = new byte[packet.len_sender % 2 != 0 ? (packet.len_sender + 1) / 2 : packet.len_sender / 2];
				Array.Copy(bytes, ++idx, packet.sender_no, 0, packet.sender_no.Length);

				idx += packet.sender_no.Length - 1;
				Utility.Swap(packet.sender_no, 0, packet.sender_no.Length);
			}

			packet.pid = bytes[++idx];
			packet.dcs = bytes[++idx];
			packet.scts = new byte[7];
			Array.Copy(bytes, ++idx, packet.scts, 0, packet.scts.Length);
			Utility.Swap(packet.scts, 0, packet.scts.Length);
			idx += packet.scts.Length - 1;
			packet.udl = bytes[++idx];

			// By default add one extra byte to the message bytes array

			packet.msg = new byte[(packet.MessageEncoding == EncodingType.GSM) ?
				GsmEncoding.GSM.GetMaxByteCount(packet.udl) : packet.udl];

			// But copy only the required stuff
			Array.Copy(bytes, ++idx, packet.msg, 0, (packet.MessageEncoding == EncodingType.GSM) ?
				GsmEncoding.GSM.GetByteCount(packet.udl) : packet.udl);

			packet.Process();
			return packet;
		}
	}

	/// <summary>
	/// SMS SUBMIT Packet
	/// </summary>
	public class SmsSubmitPacket {

		#region Private Members
		private byte len_smsc;
		private byte toa_smsc;
		private byte[] smsc_no = null;

		private byte first_octet;

		private byte tp_msg_ref;

		private byte len_dest;
		private byte toa_dest;
		private byte[] dest_no = null;

		private byte tp_pid;
		private byte tp_dcs;
		private byte[] tp_vp = null;

		private byte udl;
		private byte[] msg = null;
		#endregion

		#region Constant Stuff...
		private const int TP_RD = 2;
		private const int TP_VPF_1 = 3;
		private const int TP_VPF_2 = 4;
		private const int TP_SRR = 5;
		private const int TP_UDHI = 6;
		private const int TP_RP = 7;

		/// <summary>
		/// Maximum Length of SMS. i.e. 140 octets
		/// </summary>
		private const int MAX_LEN = 140;
		#endregion

		/// <summary>
		/// Gets or Sets whether delivery report is requested for this SMS
		/// </summary>
		/// <value>Is delivery report requested</value>
		public bool WantsDeliveryReport {
			get {
				return (SmsSubmitPacket.GetBit(this.first_octet, SmsSubmitPacket.TP_SRR) == 1) ? true : false;
			}
			set {
				SmsSubmitPacket.SetBit(ref this.first_octet, SmsSubmitPacket.TP_SRR, value ? 1 : 0);
			}
		}

		/// <summary>
		/// Gets or Sets whether to reject duplicates.
		/// </summary>
		/// <value>Whether duplicates are rejected</value>
		/// <remarks>
		/// Parameter indicating whether or not the SC shall accept an 
		/// SMS-SUBMIT for an Short Message still held in the SMS Center which has 
		/// the same <em>Message Reference</em> and the same <em>Destination Address</em> as 
		/// a previously submitted Short Message from the same <em>Originating Address</em>.
		/// </remarks>
		public bool RejectDuplicates {
			get {
				return (SmsSubmitPacket.GetBit(this.first_octet, SmsSubmitPacket.TP_RD) == 1) ? true : false;
			}
			set {
				SmsSubmitPacket.SetBit(ref this.first_octet, SmsSubmitPacket.TP_RD, value ? 1 : 0);
			}
		}

		/// <summary>
		/// Gets or Sets whether reply path exists for this SMS
		/// </summary>
		/// <value>Is Reply Path Exists</value>
		public bool ReplyPathExists {
			get {
				return (SmsSubmitPacket.GetBit(this.first_octet, SmsSubmitPacket.TP_RP) == 1) ? true : false;
			}
			set {
				SmsSubmitPacket.SetBit(ref this.first_octet, SmsSubmitPacket.TP_RP, value ? 1 : 0);
			}
		}

		/// <summary>
		/// Gets or sets whether user data exists
		/// </summary>
		/// <value>True if user data exists else false</value>
		public bool UserDataHeaderExists {
			get {
				return (SmsSubmitPacket.GetBit(this.first_octet, SmsSubmitPacket.TP_UDHI) == 1) ? true : false;
			}
			set {
				SmsSubmitPacket.SetBit(ref this.first_octet, SmsSubmitPacket.TP_UDHI, value ? 1 : 0);
			}
		}

		/// <summary>
		/// Gets or Sets Type of Validity
		/// </summary>
		/// <value>Type of Validity</value>
		public ValidityType Validity {
			get {
				return ((ValidityType)((byte)(SmsSubmitPacket.GetBit(this.first_octet,
					SmsSubmitPacket.TP_VPF_2) << 1) | SmsSubmitPacket.GetBit(this.first_octet, 
					SmsSubmitPacket.TP_VPF_1)));
			}
			set {
				byte b = (byte)value;
				SmsSubmitPacket.SetBit(ref this.first_octet, SmsSubmitPacket.TP_VPF_1, (b & 1));
				SmsSubmitPacket.SetBit(ref this.first_octet, SmsSubmitPacket.TP_VPF_2, ((b & 2) >> 1));
			}
		}

		/// <summary>
		/// Gets or Sets Destination Address
		/// </summary>
		/// <value>Destination Address</value>
		public string DestinationAddress {
			get {
				if (dest_no == null)
					return null;
				return Utility.BytesToPhone(toa_dest, len_dest, 0, dest_no);
			}
			set {
				if (value == null)
					throw new ArgumentNullException();
				string s = value;
				dest_no = Utility.PhoneToBytes(s, toa_dest, 0);
				len_dest = (byte)s.Length;
				if (s.StartsWith("+"))
					len_dest--;
			}
		}

		/// <summary>
		/// Gets or sets SMSC no.
		/// </summary>
		/// <value>SMSC No.</value>
		/// <remarks>
		/// If default SMSC has to be used, then set this property to null (or Nothing in Visual Basic)
		/// </remarks>
		public string SMSC {
			get {
				if (smsc_no == null)
					return null;
				return Utility.BytesToPhone(toa_smsc, len_smsc, 1, smsc_no);
			}
			set {
				if (value == null) {
					this.smsc_no = null;
					this.len_smsc = 0;
					return;
				}

				string s = value;
				int l = s.Length;

				if (s.StartsWith("+"))
					l--;
				smsc_no = Utility.PhoneToBytes(s, toa_smsc, 1);

				// l bytes of phone no. and 1 byte for type of address
				len_smsc = (byte)((l / 2) + 1);
			}
		}

		/// <summary>
		/// Gets or Sets Type of Address for SMSC No.
		/// </summary>
		/// <value>Type of Address SMSC No.</value>
		public byte SMSC_ToA {
			get {
				return toa_smsc;
			}
			set {
				this.toa_smsc = value;
			}
		}

		/// <summary>
		/// Gets or Sets Type of Address for Destination Address
		/// </summary>
		/// <value>Type of Address Destination Address</value>
		public byte DestinationAddress_ToA {
			get {
				return toa_dest;
			}
			set {
				this.toa_dest = value;
			}
		}

		/// <summary>
		/// Sets the expiry date/time or expiry period for this message
		/// </summary>
		/// <value>Expiry Date/Time or Expiry Period</value>
		/// <remarks>
		/// In case the <see cref="SmsSubmitPacket.Validity"/> is <see cref="ValidityType.Relative"/>, the
		/// parameter should be the <see cref="TimeSpan"/> object, else if <see cref="SmsSubmitPacket.Validity"/>
		/// is <see cref="ValidityType.Absolute"/>, this parameter should be the <see cref="DateTime"/> object.
		/// This property is not yet implemented for <strong style="color: red"><see cref="ValidityType.Enhanced"/>
		/// </strong>. This property should only be specified when <see cref="SmsSubmitPacket.Validity"/> is not
		/// set to <see cref="ValidityType.NotPresent"/>
		/// </remarks>
		public object Expires {
			get {
				switch(this.Validity) {
					case ValidityType.Absolute:
						return Utility.BytesToTimestamp(this.tp_vp);
					case ValidityType.Relative:
						return Utility.ByteToTimespan(this.tp_vp[0]);
					default:
						return null;
				}
			}
			set {
				switch (this.Validity) {
					case ValidityType.Absolute:
						this.tp_vp = Utility.TimestampToBytes((DateTime)value);
						break;
					case ValidityType.Relative:
						this.tp_vp = new byte[] {
								Utility.TimespanToByte((TimeSpan)value)
						};
						break;
					default:
						throw new NotImplementedException();
				}
			}
		}

		/// <summary>
		/// Gets or Sets the encoding of the message
		/// </summary>
		/// <value>Message Encoding</value>
		public EncodingType MessageEncoding {
			get {
				return Utility.GetEncoding(this.tp_dcs);
			}
			set {
				EncodingType prev = Utility.GetEncoding(this.tp_dcs);
				byte b = (byte)value;
				SmsSubmitPacket.SetBit(ref this.tp_dcs, 2, b & 1);
				SmsSubmitPacket.SetBit(ref this.tp_dcs, 3, (b & 2) >> 1);

				if (this.msg != null) {
					Encoding encPrev = SmsSubmitPacket.GetEncoding(prev);
					Encoding enc = SmsSubmitPacket.GetEncoding(value);
					if (encPrev == null)
						return;
					if(enc == null)
						this.msg = null;
					else {
						string s = encPrev.GetString(this.msg);
						this.msg = enc.GetBytes(s);
						this.udl = (byte)((value == EncodingType.GSM) ? s.Length : this.msg.Length);
					}
				}
			}
		}

		/// <summary>
		/// Gets or sets the message class
		/// </summary>
		/// <value>Message Class</value>
		public MessageClass MessageClass {
			get {
				return Utility.GetMessageClass(this.tp_dcs);
			}
			set {
				byte b = (byte)value;
				SmsSubmitPacket.SetBit(ref this.tp_dcs, 0, b & 1);
				SmsSubmitPacket.SetBit(ref this.tp_dcs, 1, (b & 2) >> 1);
			}
		}

		/// <summary>
		/// Gets or sets Message Text
		/// </summary>
		/// <value>Message Text</value>
		public string MessageText {
			get {
				if (this.msg == null)
					return null;

				Encoding enc = SmsSubmitPacket.GetEncoding(this.MessageEncoding);

				int ml = this.msg.Length;

				if (this.MessageEncoding == EncodingType.GSM)
					ml = GsmEncoding.GSM.GetMaxByteCount(this.udl) - GsmEncoding.GSM.GetByteCount(this.udl);

				if (enc == null)
					return null;

				string sz = enc.GetString(this.msg); //, 0, ml);

				if (this.MessageEncoding == EncodingType.GSM) {
					sz = sz.Substring(0, this.udl);
					sz = GsmHelper.Unescape(sz);
				}

				return sz;
			}
			set {
				Encoding enc = SmsSubmitPacket.GetEncoding(this.MessageEncoding);
				if (enc == null)
					throw new NotImplementedException("Support for the reserved encoding is not available");
				this.msg = enc.GetBytes(value);
				this.udl = (byte)((this.MessageEncoding == EncodingType.GSM) ? GsmEncoding.GSM.GetMaxCharCount(this.msg.Length): this.msg.Length);
			}
		}

		/// <summary>
		/// Gets message text length according to the current encoding
		/// </summary>
		/// <value>Message Text Length</value>
		/// <remarks>
		/// If current message encoding is GSM, then this is the no. of septets, else if it is an UCS2 encoded
		/// or an Octet encoded Message it is the no. of octets or no. of bytes.
		/// </remarks>
		public int MessageLength {
			get {
				return this.udl;
			}
		}

		/// <summary>
		/// Gets or sets Message Reference Number
		/// </summary>
		/// <value>Message Reference Number</value>
		/// <remarks>
		/// The default value is 0 which means that this number is set by phone
		/// </remarks>
		public byte MessageRefNumber {
			get {
				return this.tp_msg_ref;
			}
			set {
				this.tp_msg_ref = value;
			}
		}

		#region Private Statix Methods
		/// <summary>
		/// Gets octet's bit
		/// </summary>
		/// <param name="b">Octet</param>
		/// <param name="bit">Position of the bit</param>
		/// <returns>1 if bit is set else 0</returns>
		private static byte GetBit(byte b, int bit) {
			return ((byte)((b & (1 << bit)) >> bit));
		}

		/// <summary>
		/// Sets/Resets an octet's bit
		/// </summary>
		/// <param name="b">Octet</param>
		/// <param name="bit">Position of the bit</param>
		/// <param name="v">1 if bit is to be set else 0</param>
		private static void SetBit(ref byte b, int bit, int v) {
			byte x = (byte)((v == 0) ? ~(1 << bit) : (1 << bit));
			if (v == 0)
				b &= x;
			else
				b |= x;
		}

		/// <summary>
		/// Gets <see cref="Encoding"/> instance for the encoding type
		/// </summary>
		/// <param name="t"><see cref="EncodingType"/> for which <see cref="Encoding"/> has to be returned</param>
		/// <returns>A <see cref="Encoding"/> instance</returns>
		/// <remarks>
		/// In case <paramref name="t"/> is <see cref="EncodingType.Reserved"/>, the return value is null.
		/// </remarks>
		private static Encoding GetEncoding(EncodingType t) {
			Encoding enc = null;
			switch (t) {
				case EncodingType.GSM:
					enc = GsmEncoding.GSM;
					break;
				case EncodingType.OCTET:
					enc = Encoding.ASCII;
					break;
				case EncodingType.UCS2:
					enc = Encoding.BigEndianUnicode;
					break;
				case EncodingType.Reserved:
					enc = null;
					break;
			}

			return enc;
		}
		#endregion

		/// <summary>
		/// Constructs a new instance of the <see cref="SmsSubmitPacket"/>
		/// </summary>
		public SmsSubmitPacket() {
			this.len_smsc = 0;
			this.len_dest = 0;
			this.toa_smsc = 0;
			this.toa_dest = 0;

			this.tp_msg_ref = 0;

			// An 7bit GSM Class 1 Message
			this.tp_dcs = 0x11;

			// An SMS
			this.tp_pid = 0;

			this.first_octet = 1;

			this.udl = 0;
		}

		/// <summary>
		/// Parses
		/// </summary>
		/// <param name="bytes">Array of bytes</param>
		/// <returns>An instance of <see cref="SmsSubmitPacket"/> if parsing is successful</returns>
		public static SmsSubmitPacket Parse(byte[] bytes) {
			SmsSubmitPacket packet = new SmsSubmitPacket();
			int idx = 0;

			packet.len_smsc = bytes[idx++];

			if (packet.len_smsc != 0) {
				packet.toa_smsc = bytes[idx++];
				packet.smsc_no = new byte[packet.len_smsc - 1];
				Array.Copy(bytes, idx++, packet.smsc_no, 0, packet.smsc_no.Length);
				idx += packet.smsc_no.Length - 1;
				Utility.Swap(packet.smsc_no, 0, packet.smsc_no.Length);
			}

			packet.first_octet = bytes[idx++];
			packet.tp_msg_ref = bytes[idx++];
			packet.len_dest = bytes[idx++];

			if (packet.len_dest != 0) {
				packet.toa_dest = bytes[idx++];
				packet.dest_no = new byte[packet.len_dest % 2 != 0 ? (packet.len_dest + 1) / 2 : packet.len_dest / 2];
				Array.Copy(bytes, idx++, packet.dest_no, 0, packet.dest_no.Length);

				idx += packet.dest_no.Length - 1;
				Utility.Swap(packet.dest_no, 0, packet.dest_no.Length);
			}

			packet.tp_pid = bytes[idx++];
			packet.tp_dcs = bytes[idx++];

			if (packet.Validity != ValidityType.NotPresent) {
				switch (packet.Validity) {
					case ValidityType.Relative:
						packet.tp_vp = new byte[1];
						packet.tp_vp[0] = bytes[idx++];
						break;
					case ValidityType.Absolute:
						packet.tp_vp = new byte[7];
						Array.Copy(bytes, idx++, packet.tp_vp, 0, packet.tp_vp.Length);
						Utility.Swap(packet.tp_vp, 0, packet.tp_vp.Length);
						idx += packet.tp_vp.Length - 1;
						break;
					case ValidityType.Enhanced:
						packet.tp_vp = new byte[7];
						Array.Copy(bytes, idx++, packet.tp_vp, 0, packet.tp_vp.Length);
						idx += packet.tp_vp.Length - 1;
						break;
				}
			}

			packet.udl = bytes[idx++];
			packet.msg = new byte[(packet.MessageEncoding == EncodingType.GSM) ?
				GsmEncoding.GSM.GetByteCount(packet.udl) : packet.udl];

			Array.Copy(bytes, idx++, packet.msg, 0, (packet.MessageEncoding == EncodingType.GSM) ?
				GsmEncoding.GSM.GetByteCount(packet.udl) : packet.udl);

			return packet;
		}

		/// <summary>
		/// Gets this SM as array of bytes
		/// </summary>
		/// <returns>An array of bytes</returns>
		public byte[] GetBytes() {
			int ln_dest = (len_dest % 2) == 0 ? (len_dest / 2) : ((len_dest / 2) + 1);
			int len = 1 + len_smsc + 1 + 1 + 1 + 1 + ln_dest + 1 + 1 +
				(this.Validity == ValidityType.NotPresent ? 0 : this.tp_vp.Length) + 1 +
				this.msg.Length;
//				((this.MessageEncoding == EncodingType.GSM) ? GsmEncoding.GSM.GetCharCount(this.msg) : this.msg.Length);
			byte[] bytes = new byte[len];
			int i = 0;

			bytes[i++] = this.len_smsc;
			if (this.len_smsc > 0) {
				bytes[i++] = this.toa_smsc;

				Utility.Swap(this.smsc_no, 0, this.smsc_no.Length);
				Array.Copy(this.smsc_no, 0, bytes, i++, len_smsc - 1);
				Utility.Swap(this.smsc_no, 0, this.smsc_no.Length);

				i += len_smsc - 2;
			}
			bytes[i++] = this.first_octet;
			bytes[i++] = this.tp_msg_ref;
			bytes[i++] = this.len_dest;
			bytes[i++] = this.toa_dest;

			Utility.Swap(this.dest_no, 0, this.dest_no.Length);
			Array.Copy(this.dest_no, 0, bytes, i++, ln_dest);
			Utility.Swap(this.dest_no, 0, this.dest_no.Length);

			i += ln_dest - 1;

			bytes[i++] = this.tp_pid;
			bytes[i++] = this.tp_dcs;

			if (this.Validity != ValidityType.NotPresent) {
				if (this.Validity == ValidityType.Absolute)
					Utility.Swap(this.tp_vp, 0, this.tp_vp.Length);

				Array.Copy(this.tp_vp, 0, bytes, i++, this.tp_vp.Length);

				if (this.Validity == ValidityType.Absolute)
					Utility.Swap(this.tp_vp, 0, this.tp_vp.Length);

				i += this.tp_vp.Length - 1;
			}

			bytes[i++] = this.udl;
			Array.Copy(this.msg, 0, bytes, i++, this.msg.Length);

			return bytes;
		}
	}

	/// <summary>
	/// SMS STATUS-REPORT Packet
	/// </summary>
	public class SmsStatusReportPacket {

		#region Private members
		private byte len_smsc;
		private byte toa_smsc;
		private byte[] smsc_no = null;

		private byte first_octet;

		private byte len_sender;
		private byte toa_sender;
		private byte[] sender_no = null;

		private byte[] scts = null;

		private string sender = null;
		private string smsc = null;

		private DateTime timestamp;

		private bool delivered = false;
		private byte[] delivery_scts = null;
		private DateTime delivery_timestamp;
		#endregion

		/// <summary>
		/// Gets Sender Number
		/// </summary>
		/// <value>Sender Number</value>
		public string OriginatingAddress {
			get {
				return sender;
			}
		}

		/// <summary>
		/// Gets Type of Address of SMSC
		/// </summary>
		/// <value>Type of Address</value>
		public byte SMSC_ToA {
			get {
				return toa_smsc;
			}
		}

		/// <summary>
		/// Gets Type of Address of Sender
		/// </summary>
		/// <value>Type of Address</value>
		public byte OriginatingAddress_ToA {
			get {
				return toa_sender;
			}
		}

		/// <summary>
		/// Whether the delivery reports indicates a failure or success of 
		/// message delivery
		/// </summary>
		/// <value>True if successful</value>
		public bool Delivered {
			get {
				return delivered;
			}
		}

		/// <summary>
		/// Gets SMSC (Short Message Service Center or SMS Center) No.
		/// </summary>
		/// <value>SMS Center No.</value>
		public string SMSC {
			get {
				return smsc;
			}
		}

		/// <summary>
		/// Gets timestamp on this message
		/// </summary>
		/// <value>Timestamp on this message</value>
		public DateTime Timestamp {
			get {
				return timestamp;
			}
		}

		/// <summary>
		/// Gets UTC Offset for the timestamp's timezone
		/// </summary>
		/// <value>UTC Offset of the timestamp</value>
		public TimeSpan TimeZone {
			get {
				return Utility.QuartersBcdToUtc(this.scts[6]);
			}
		}

		/// <summary>
		/// Gets delivery timestamp on this message
		/// </summary>
		/// <value>Delivery timestamp on this message</value>
		public DateTime DeliveryTimestamp {
			get {
				return delivery_timestamp;
			}
		}

		/// <summary>
		/// Gets UTC Offset for the delivery timestamp's timezone
		/// </summary>
		/// <value>UTC Offset of the delivery timestamp</value>
		public TimeSpan DeliveryTimeZone {
			get {
				return Utility.QuartersBcdToUtc(this.delivery_scts[6]);
			}
		}

		/// <summary>
		/// Process this SMS packet
		/// </summary>
		private void Process() {

			timestamp = Utility.BytesToTimestamp(this.scts);
			delivery_timestamp = Utility.BytesToTimestamp(this.delivery_scts);

			this.sender = Utility.BytesToPhone(this.toa_sender, this.len_sender, 0, this.sender_no);
			this.smsc = Utility.BytesToPhone(this.toa_smsc, this.len_smsc, 1, this.smsc_no);
		}

		/// <summary>
		/// Creates an instance of the <see cref="SmsStatusReportPacket"/> by parsing the
		/// SMS STATUS REPORT packet bytes.
		/// </summary>
		/// <param name="bytes">SMS STATUS REPORT packet bytes</param>
		/// <returns>An instance of <see cref="SmsStatusReportPacket"/> if successful,
		/// otherwise an exception is thrown</returns>
		public static SmsStatusReportPacket Parse(byte[] bytes) {
			SmsStatusReportPacket packet = new SmsStatusReportPacket();
			int idx = 0;

			packet.len_smsc = bytes[idx++];
			if (packet.len_smsc != 0) {
				packet.toa_smsc = bytes[idx++];
				packet.smsc_no = new byte[packet.len_smsc - 1];
				Array.Copy(bytes, idx++, packet.smsc_no, 0, packet.smsc_no.Length);
				idx += packet.smsc_no.Length - 1;
				Utility.Swap(packet.smsc_no, 0, packet.smsc_no.Length);
			}

			packet.first_octet = bytes[idx++];

			// Skipped one field I don't know what it does. ;)
			idx++;

			packet.len_sender = bytes[idx++];
			packet.toa_sender = bytes[idx++];
			packet.sender_no = new byte[(packet.len_sender % 2) != 0 ? (packet.len_sender + 1) / 2 : packet.len_sender / 2];
			Array.Copy(bytes, idx++, packet.sender_no, 0, packet.sender_no.Length);
			idx += packet.sender_no.Length - 1;
			Utility.Swap(packet.sender_no, 0, packet.sender_no.Length);

			packet.scts = new byte[7];
			Array.Copy(bytes, idx++, packet.scts, 0, packet.scts.Length);
			idx += packet.scts.Length - 1;
			Utility.Swap(packet.scts, 0, packet.scts.Length);

			packet.delivery_scts = new byte[7];
			Array.Copy(bytes, idx++, packet.delivery_scts, 0, packet.delivery_scts.Length);
			idx += packet.delivery_scts.Length - 1;
			Utility.Swap(packet.delivery_scts, 0, packet.delivery_scts.Length);

			packet.delivered = ((bytes[idx++] == 0) ? true : false);

			packet.Process();

			return packet;
		}
	}
	#endregion
}
 