﻿#region Using directives

using System;
using System.Collections.Generic;
using System.Text;

#endregion

namespace MonsieurWJ.LibXMS2 {

	/// <summary>
	/// Utility Class
	/// </summary>
	public class Utility {

		/// <summary>
		/// Not instantiable
		/// </summary>
		private Utility() {
		}

		public static string GetReply(string text, string ls) {
			return "OK";
		}

		public static string GetProvider(string ati_reply, string ls) {
			// "ATI3\r\n\r\nNokia 6610\r\n\r\nOK\r\n"
			if (ati_reply.IndexOf("OK") < 0) {
				return null;
			}
			string[] cmds = ati_reply.Split(new char[] { '\r', '\n' }, StringSplitOptions.RemoveEmptyEntries);

			return cmds[1];
		}

		#region Decimal - BCD Conversion Routines
		/// <summary>
		/// BCD to Decimal conversion routine
		/// </summary>
		/// <remarks>
		/// If this method encounter any hexadecimal digit >= A or decimal equivalent > 9
		/// it changes that digit to zero in the result. Only for the 1s (ones) place in
		/// hexa decimal number
		/// </remarks>
		/// <param name="val">BCD Value to convert</param>
		/// <returns>Converted decimal number</returns>
		public static byte BcdToDecimal(byte val) {
			byte m;
			return (byte)(((byte)((val & 0xf0) >> 4) * 10) + ((m = (byte)(val & ((byte)0x0f))) > ((byte)0x9) ? ((byte)0) : m));
		}

		/// <summary>
		/// Decimal to BCD conversion routine
		/// </summary>
		/// <param name="val">
		/// If this method encountered any decimal number > 99 then it return 0.
		/// </param>
		/// <returns>Converted BCD number</returns>
		public static byte DecimalToBcd(byte val) {
			return (val > 99) ? (byte)0 : (byte)((((val / 10) << 4)) | ((val % 10)));
		}
		#endregion

		#region Phone Number - Bytes Conversion Routines
		/// <summary>
		/// Converts BCD bytes to the phone no.
		/// </summary>
		/// <param name="toa">Type of Address byte</param>
		/// <param name="len">Length of the number as indicated in the packet</param>
		/// <param name="flag">Flag which tells whether this is SMSC no, or sender
		/// /reciever no. 0 for Sender and Reciever while all other is for SMSC no.</param>
		/// <param name="no">Bytes of the number as fetched from packet. This function
		/// assumes that number bytes had swapped after fetching</param>
		/// <returns>A Phone Number</returns>
		public static string BytesToPhone(byte toa, byte len, byte flag, byte[] no) {

			string number = "";
			byte num;

			if (toa == (byte)AddressType.GSM) {

				// Number is in GSM Encoding
				// And is in swapped state since we already swapped
				// the array while parsing, so we're assuming that 
				// array is already swapped so reswapping it.
				Utility.Swap(no, 0, no.Length);
				number = new string(GsmEncoding.GSM.GetChars(no));

				// Reswap it, to return to its original state
				Utility.Swap(no, 0, no.Length);
			} else {

				for (int i = 0; i < no.Length; i++)
					number += "" + ((num = Utility.BcdToDecimal(no[i])) < 10 ? "0" + num : "" + num);

				// Remove the last 'F' char if the length is odd no.
				if (flag == 0) {
					if ((len % 2) != 0) {
						number = number.Substring(0, number.Length - 1);
					}
				}
				if (toa == (byte)AddressType.International)
					number = "+" + number;
			}

			return number;
		}

		/// <summary>
		/// Converts phone number to the BCD bytes
		/// </summary>
		/// <param name="toa">Type of Address byte</param>
		/// <param name="flag">Flag which tells whether this is SMSC no, or sender
		/// /reciever no. 0 for Sender and Reciever No. while all other is SMSC no.</param>
		/// <param name="phone">Bytes of the phone number as fetched from packet. This function
		/// assumes that number bytes had swapped after fetching</param>
		/// <returns>Array of Bytes</returns>
		public static byte[] PhoneToBytes(string phone, byte toa, byte flag) {
			string p = phone;
			byte[] b = null;
			bool odd = false;

			if (p.StartsWith("+"))
				p = phone.Substring(1);

			odd = ((p.Length % 2) == 1);

			if (toa == (byte)AddressType.GSM) {
				b = GsmEncoding.GSM.GetBytes(p);
				Utility.Swap(b, 0, b.Length);
			} else {
				if (odd)
					p += "0";
				b = HexEncoding.HEX.GetBytes(p);
				if (odd)
					b[b.Length - 1] |= 0xf;
			}
			return b;
		}
		#endregion

		#region Bytes - Timestamp Conversion Rountines
		/// <summary>
		/// Gets timestamp from the array of bytes.
		/// </summary>
		/// <remarks><see cref="DateTimeKind"/> is in Local Time Zone.
		/// The nibbles of the bytes are expected to be already swapped.
		/// </remarks>
		/// <param name="bytes">Array of bytes. (SCTS)</param>
		/// <returns>A <see cref="DateTime"/> instance</returns>
		public static DateTime BytesToTimestamp(byte[] bytes) {
			return new DateTime(2000 + Utility.BcdToDecimal(bytes[0]),
				Utility.BcdToDecimal(bytes[1]), Utility.BcdToDecimal(bytes[2]),
				Utility.BcdToDecimal(bytes[3]), Utility.BcdToDecimal(bytes[4]),
				Utility.BcdToDecimal(bytes[5]), DateTimeKind.Local);
		}

		/// <summary>
		/// Converts Timestamp to SMSC Timestamp Bytes
		/// </summary>
		/// <remarks>
		/// This method already initializes timezone, if the <see cref="DateTime.Kind"/> of <paramref name="time"/>
		/// is <see cref="DateTimeKind.Local"/> with the current time zone's UTC offset in quarters of an hour else 0.
		/// The nibbles of the bytes have to be swapped.
		/// </remarks>
		/// <param name="time">Time to be converted</param>
		/// <returns>An array of usually 7 bytes</returns>
		public static byte[] TimestampToBytes(DateTime time) {
			return new byte[] {
				Utility.DecimalToBcd((byte)(time.Year-2000)), 
				Utility.DecimalToBcd((byte)time.Month), 
				Utility.DecimalToBcd((byte)time.Day),
				Utility.DecimalToBcd((byte)time.Hour),
				Utility.DecimalToBcd((byte)time.Minute),
				Utility.DecimalToBcd((byte)time.Second),
				((time.Kind == DateTimeKind.Local) ? Utility.UtcToQuartersBcd(
					TimeZone.CurrentTimeZone.GetUtcOffset(DateTime.Now)):(byte)0)
			};
		}
		#endregion

		#region UTC - Quarters BCD Conversion Routines
		/// <summary>
		/// Converts an UTC offset (formerly GMT equivalent) to multiples of quarter of an hour
		/// </summary>
		/// <param name="zone">UTC Offset</param>
		/// <returns>Converted TimeZone UTC offset into multiples of quarter
		/// of an hour</returns>
		public static byte UtcToQuartersBcd(TimeSpan utc) {
			byte b = Utility.DecimalToBcd((byte)(Math.Abs(utc.Hours * 4) + Math.Abs(utc.Minutes / 15)));
			return (((utc.Hours < 0) || (utc.Minutes < 0)) ? ((byte)(b | 0x80)) : b);
		}

		/// <summary>
		/// Converts multiples of quarter of an hour to an UTC offset (formerly GMT equivalent).
		/// </summary>
		/// <param name="b">Multiples of Quarter of an hour</param>
		/// <returns>A <see cref="TimeSpan"/> instance representing an UTC offset</returns>
		public static TimeSpan QuartersBcdToUtc(byte b) {
			int sgn = (((b & 0x80) == 0x80) ? -1 : 1);
			int t = Utility.BcdToDecimal((byte)(b & 0x7f));
			return new TimeSpan((t / 4) * sgn, (t % 4) * 15 * sgn, 0);
		}
		#endregion

		#region TimeSpan - Byte Conversion Routines
		/// <summary>
		/// Converts byte to <see cref="TimeSpan"/> according to the validity period specification
		/// </summary>
		/// <param name="p">Byte to be converted</param>
		/// <returns>A <see cref="TimeSpan"/> instance</returns>
		public static TimeSpan ByteToTimespan(byte p) {
			TimeSpan span = TimeSpan.Zero;

			if (p >= 0 && p <= 143)
				span = new TimeSpan(0, (p + 1) * 5, 0);
			else if (p >= 144 && p <= 167)
				span = new TimeSpan(12, (p - 143) * 30, 0);
			else if (p >= 168 && p <= 196)
				span = new TimeSpan((p - 166), 0, 0, 0);
			else if (p >= 197 && p <= 255)
				span = new TimeSpan((p - 192) * 7, 0, 0, 0);

			return span;
		}

		/// <summary>
		/// Converts <see cref="TimeSpan"/> to byte according to the validity period specification
		/// </summary>
		/// <param name="span"><see cref="TimeSpan"/> to be converted</param>
		/// <returns>A byte</returns>
		public static byte TimespanToByte(TimeSpan span) {
			byte b = 0;
			int i;

			if (span.TotalMinutes > 0 && span.TotalMinutes <= 720.0) {
				i = (int)span.TotalMinutes;

				// Round off to next 5
				if (i % 5 != 0)
					i += (i % 5);
				i /= 5;
				b = (byte)(i-1);
			} else if (span.TotalHours > 12.0 && span.TotalHours <= 24.0) {
				i = (int)span.TotalMinutes;

				i -= 720;
				if (i % 30 != 0)
					i += (i % 30);
				i /= 30;
				b = (byte)(i + 143);
			} else if (span.TotalDays > 1.0 && span.TotalDays <= 30.0) {
				i = (int)Math.Ceiling(span.TotalDays);

				b = (byte)(i + 166);
			} else if (span.TotalDays > 30.0) {
				i = (int)Math.Ceiling(span.TotalDays);
				i -= 30;

				if (i % 7 != 0)
					i = (i/7) +1;
				else
					i /= 7;

				b = (byte)(i + 196);
			}

			return b;
		}
		#endregion
	
		#region Other Helper Stuff
		/// <summary>
		/// Gets alphabet encoding represented by the <paramref name="dcs"/>.
		/// </summary>
		/// <param name="dcs">TP-DCS field of the SMS Packet</param>
		/// <returns>An <see cref="EncodingType"/> instance representing
		/// the encoding</returns>
		public static EncodingType GetEncoding(byte dcs) {
			// 6th and 7th bits are 0
			if ((dcs & 0xC0) == 0) {
				return ((EncodingType)((dcs & 0xC) >> 2));
/*				{
					case 0:
						// 00
						return EncodingType.GSM;
					case 1:
						// 01
						return EncodingType.OCTET;
					case 2:
						// 10
						return EncodingType.UCS2;
					case 3:
						return EncodingType.Reserved;
				} */
			} else if ((dcs & 0xF0) == 0xf0) {
				if ((dcs & 4) == 0)
					return EncodingType.GSM;
				else
					return EncodingType.OCTET;
			}
			// Fallback to default encoding
			return EncodingType.GSM;
		}

		/// <summary>
		/// Gets message class represented by the <paramref name="dcs"/>.
		/// </summary>
		/// <remarks>
		/// In case the TP-DCS fields returns no message class a 
		/// <see cref="MessageClass.Class1"/> will be returned.
		/// </remarks>
		/// <param name="dcs">TP-DCS field of the SMS Packet</param>
		/// <returns>An <see cref="MessageClass"/> instance representing
		/// the message class.</returns>
		public static MessageClass GetMessageClass(byte dcs) {
			if ((dcs & 0xC0) == 0) {
				if ((dcs & 0x10) == 0x10) {
					return ((MessageClass)((dcs & 0x3))); /* {
						case 0:
							// 00
							return MessageClass.Class0;
						case 1:
							// 01
							return MessageClass.Class1;
						case 2:
							// 10
							return MessageClass.Class2;
						case 3:
							// 11
							return MessageClass.Class3;
					} */
				}
				//Console.WriteLine("Message Class: {0}", dcs);
				return MessageClass.Class1;
			} else if ((dcs & 0xf0) == 0xf0) {
				switch ((dcs & 0x3)) {
					case 0:
						// 00
						return MessageClass.Class0;
					case 1:
						// 01
						return MessageClass.Class1;
					case 2:
						// 10
						return MessageClass.Class2;
					case 3:
						// 11
						return MessageClass.Class3;
				}
				Console.WriteLine("Message Class: {0}", dcs);
				return MessageClass.Class1;
			} else
				return MessageClass.Class1;
		}

		/// <summary>
		/// Swaps the nibbles of the elements of the given array of bytes
		/// </summary>
		/// <param name="array">Array of bytes</param>
		/// <param name="src">Source Index</param>
		/// <param name="len">Length</param>
		public static void Swap(byte[] array, int src, int len) {
			for (int i = src; i < src + len; i++)
				array[i] = (byte)(((array[i] & 0xf0) >> 4) | ((array[i] & 0xf) << 4));
		}

		/// <summary>
		/// Tells whether string can be encoded in Unicode characters only.
		/// </summary>
		/// <param name="text">String to be tested</param>
		/// <returns>Return true if string can be encoded in Unicode else false</returns>
		/// <remarks>
		/// This method checks by comparing each character(s) Unicode value to be less than
		/// 128. And if the value is less than 128, then the character can be safely encoded
		/// in other (single byte or less than 8 bits) encodings like ASCII, PCCP437 etc.
		/// But if it is greater than or equal 128 this means that string has to be encoded in
		/// Unicode.
		/// </remarks>
		public static bool IsStringUnicode(string text) {
			char[] ch = text.ToCharArray();
			foreach (int c in ch) {
				if (!(c >= 32 && c <= 255))
					return true;
			}
			return false;
		}
		#endregion
	}
}
